from itertools import chain

from pdf417.data import CHARACTERS_LOOKUP, SWITCH_CODES, Submode
from pdf417.util import switch_base, to_base, chunks


# -- Number compaction mode ----------------------------------------------------

def compact_numbers(data):
    """Encodes data into code words using the Numbers compaction mode.

    Can encode: Digits 0-9, ASCII
    Rate compaction: 2.9 bytes per code word
    """

    def compact_chunk(chunk):
        number = "".join([chr(x) for x in chunk])
        value = int("1" + number)
        return to_base(value, 900)

    compacted_chunks = [compact_chunk(chunk) for chunk in chunks(data, size=44)]

    return chain(*compacted_chunks)


# -- Text compaction mode ------------------------------------------------------

def compact_text_interim(data):
    def exists_in_submode(char, submode):
        return char in CHARACTERS_LOOKUP and \
               submode in CHARACTERS_LOOKUP[char]

    def get_submode(char):
        if char not in CHARACTERS_LOOKUP:
            raise ValueError("Cannot encode char: {}".format(char))

        submodes = CHARACTERS_LOOKUP[char].keys()

        preference = [Submode.LOWER, Submode.UPPER, Submode.MIXED,
                      Submode.PUNCT]

        for submode in preference:
            if submode in submodes:
                return submode

        raise ValueError("Cannot encode char: {}".format(char))

    # By default, encoding starts with uppercase submode
    submode = Submode.UPPER

    codes = []

    for char in data:
        # Do we need to switch submode?
        # TODO: use one-character switches
        if not exists_in_submode(char, submode):
            prev_submode = submode
            submode = get_submode(char)

            switch_codes = SWITCH_CODES[prev_submode][submode]
            codes.extend(switch_codes)

        codes.append(CHARACTERS_LOOKUP[char][submode])

    return codes


def compact_text(data):
    """Encodes data into code words using the Text compaction mode.

    Can encode: ASCII 9, 10, 13 and 32-126
    Rate compaction: 2 bytes per code word
    """

    # Since each code word consists of 2 characters, a padding value is
    # needed when encoding a single character. 29 is used as padding because
    # it's a switch in all 4 submodes, and doesn't add any data.

    PADDING_INTERIM_CODE = 29

    def compact_chunk(chunk):
        if len(chunk) == 1:
            chunk.append(PADDING_INTERIM_CODE)

        return 30 * chunk[0] + chunk[1]

    interim_codes = compact_text_interim(data)

    return [compact_chunk(chunk) for chunk in chunks(interim_codes, 2)]


# -- Bytes compaction mode -----------------------------------------------------

def compact_bytes(data):
    """Encodes data into code words using the Byte compaction mode.

    Can encode: ASCII 0 to 255
    Rate compaction: 1.2 byte per code word
    """

    def compact_chunk(chunk):
        return compact_full_chunk(chunk) if len(chunk) == 6 \
            else compact_incomplete_chunk(chunk)

    def compact_full_chunk(chunk):
        """Encodes a chunk consisting of exactly 6 bytes.

        The chunk is encoded to 5 code words by changing
        the base from 256 to 900.
        """

        digits = [i for i in chunk]
        return switch_base(digits, 256, 900)

    def compact_incomplete_chunk(chunk):
        """Encodes a chunk consisting of less than 6 bytes.

        The chunk is encoded to 6 code words leaving the base unchanged.
        """
        return [i for i in chunk]

    compacted_chunks = [compact_chunk(chunk) for chunk in chunks(data, size=6)]

    return chain(*compacted_chunks)


# -- Bringing it all together --------------------------------------------------

def compact(data, numeric_compaction=True):
    """Encodes given data into an array of PDF417 code words."""

    def compact_chunks(chunks):
        compacted_chunks = [compact_chunk(ordinal, *args) for ordinal, args in
                            enumerate(chunks)]

        return chain(*compacted_chunks)

    def compact_chunk(ordinal, chunk, compact_fn):
        code_words = []

        # Add the switch code if required
        add_switch_code = ordinal > 0 or compact_fn != compact_text
        if add_switch_code:
            code_words.append(get_switch_code(compact_fn, chunk))

        code_words.extend(compact_fn(chunk))

        return code_words

    def split_to_chunks(data):
        """Splits a string into chunks which can be encoded with the
        same encoder.

        Implemented as a generator which yields chunks and
        the appropriate encoder.

        TODO: Currently always switches to the best encoder, even if it's just
        for one character, consider a better algorithm.
        """

        # Default compaction mode is Text
        # (does not require an initial switch code)
        function = compact_text
        chunk = []

        for char in data:
            new_function = get_optimal_compactor_fn(char)
            if function != new_function:
                if chunk:
                    yield chunk, function

                chunk = []
                function = new_function
            chunk.append(char)

        if chunk:
            yield chunk, function

    def compact_equals(chunks):
        prev_chunk = None
        prev_fn = None

        for chunk, fn in iter(chunks):
            if prev_chunk is None:
                prev_chunk, prev_fn = chunk, fn
                continue

            if prev_fn == fn:
                prev_chunk.extend(chunk)
                continue

            yield prev_chunk, prev_fn
            prev_chunk, prev_fn = chunk, fn

        if prev_chunk:
            yield prev_chunk, prev_fn

    def replace_short_numbers(chunks):
        prev_chunk = None
        prev_fn = None

        for chunk, fn in iter(chunks):
            if prev_chunk is None:
                prev_chunk, prev_fn = chunk, fn
                continue

            if len(chunk) < 13 and fn is compact_numbers:
                if prev_fn is compact_text:
                    yield prev_chunk + chunk, prev_fn
                    prev_chunk, prev_fn = None, None
                else:
                    yield prev_chunk, prev_fn
                    prev_chunk, prev_fn = chunk, compact_text
            else:
                yield prev_chunk, prev_fn
                prev_chunk, prev_fn = chunk, fn

        if prev_chunk:
            yield prev_chunk, prev_fn

    chunks = split_to_chunks(data)
    if numeric_compaction:
        chunks = compact_equals(replace_short_numbers(chunks))

    return compact_chunks(chunks)


def get_optimal_compactor_fn(char):
    if 48 <= char <= 57:
        return compact_numbers

    if char in CHARACTERS_LOOKUP:
        return compact_text

    return compact_bytes


def get_switch_code(compact_fn, data):
    TEXT = 900
    BYTES = 901
    BYTES_ALT = 924
    NUMBERS = 902

    if compact_fn == compact_text:
        return TEXT

    if compact_fn == compact_bytes:
        return BYTES_ALT if len(data) % 6 == 0 else BYTES

    if compact_fn == compact_numbers:
        return NUMBERS

    assert False, "Nonexistant compaction function"



import logging
import sys

from argparse import ArgumentParser

from pdf417.encoding import encode
from pdf417.rendering import render_image


def encode_data(args):
    if not args.text:
        logging.info("Waiting data on stdin.")
        args.text = sys.stdin.read()

    codes = encode(
        args.text,
        columns=args.columns,
        security_level=args.security_level,
        encoding=args.encoding,
        numeric_compaction=args.numeric_compaction,
    )

    image = render_image(
        codes,
        scale=args.scale,
        ratio=args.ratio,
        padding=args.padding,
        fg_color=args.fg_color,
        bg_color=args.bg_color,
    )

    if args.output:
        image.save(args.output)
    else:
        image.show()


parser = ArgumentParser(epilog="https://github.com/mosquito/pdf417",
                        description="Generate a bar code from given input")

parser.set_defaults(func=lambda *_: parser.print_help())


subparsers = parser.add_subparsers(help="Encodes text to barcode")
subparser = subparsers.add_parser("encode")

subparser.add_argument("text", type=str, nargs="?",
                       help="Text or data to encode. "
                            "Alternatively data can be piped in.")

subparser.add_argument("-c", "--columns", type=int, default=6,
                       help="The number of columns (default is 6).")

subparser.add_argument("-l", "--security-level", default=2, type=int,
                       help="Security level (default is 2).")

subparser.add_argument("-e", "--encoding", type=str, default='utf-8',
                       help="Character encoding used to decode "
                            "input (default is utf-8).")

subparser.add_argument("-n", "--numeric_compaction", action='store_true',
                       help="This mode can pack almost 3 digits "
                            "(2.93) info a symbol character")

subparser.add_argument("-s", "--scale", default=3, type=int,
                       help="Module width in pixels (default is 3).")

subparser.add_argument("-r", "--ratio", dest="ratio", type=int, default=3,
                       help="Module height to width ratio (default is 3).")

subparser.add_argument("-p", "--padding", default=20, type=int,
                       help="Image padding in pixels (default is 20).")

subparser.add_argument("-f", "--foreground-color", dest="fg_color", type=str,
                       help="Foreground color in hex (default is '#000000').",
                       default="#000000")

subparser.add_argument("-b", "--background-color", dest="bg_color", type=str,
                       help="Foreground color in hex (default is '#FFFFFF'.",
                       default="#FFFFFF")

subparser.add_argument("-o", "--output", dest="output", type=str,
                       help="Target file (if not given, "
                            "will just show the barcode).")
subparser.set_defaults(func=encode_data)


def entrypoint():
    logging.basicConfig(
        format='\033[91m%(message)s\033[0m',
        level=logging.INFO
    )

    arguments = parser.parse_args()
    try:
        arguments.func(arguments)
    except Exception as e:
        logging.exception("%r", e)
        return 10

    return 0


def main():
    exit(entrypoint())


if __name__ == '__main__':
    main()



from __future__ import division
from builtins import bytes, str


def from_base(digits, base):
    return sum(
        v * (base ** (len(digits) - k - 1)) for k, v in enumerate(digits)
    )


def to_base(value, base):
    digits = []

    while value > 0:
        digits.insert(0, value % base)
        value //= base

    return digits


def switch_base(digits, source_base, target_base):
    return to_base(from_base(digits, source_base), target_base)


def chunks(data, size):
    """Generator which chunks data into 6 bytes batches"""
    for i in range(0, len(data), size):
        yield data[i:i + size]


def to_bytes(input, encoding='utf-8'):
    if isinstance(input, bytes):
        return bytes(input)

    if isinstance(input, str):
        return bytes(input, encoding)

    raise ValueError("Invalid input, expected string or bytes")
